"""Dataset Format handling"""

from __future__ import annotations

import abc
from typing import IO

import pandas as pd
import pyarrow as pa
from base64io import Base64IO
from multimethod import multimethod
from pyarrow import RecordBatchFileWriter

from arakawa import optional_libs as opt
from arakawa.types import ARROW_EXT, ARROW_MIMETYPE, MIME

from .df_processor import obj_to_str, process_df, str_to_arrow_str

PathOrFile = str | IO | Base64IO


def write_table(table: pa.Table, sink: PathOrFile):
    """Write an arrow table to a file"""
    writer = RecordBatchFileWriter(sink, table.schema)
    writer.write(table)
    writer.close()


class DFFormatter(abc.ABC):
    # TODO - tie to mimetypes lib
    content_type: MIME
    ext: str
    enum: str

    @staticmethod
    @abc.abstractmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):
        pass


DFFormatterCls = type[DFFormatter]


class ArrowFormat(DFFormatter):
    content_type = ARROW_MIMETYPE
    ext = ARROW_EXT
    enum = "ARROW"

    @staticmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        df = pa.ipc.open_file(fn).read_pandas()
        # NOTE - need to convert categories from object to string https://github.com/apache/arrow/issues/33070
        obj_to_str(df)
        str_to_arrow_str(df)
        return df

    @multimethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):  # type: ignore # noqa: N805
        df = process_df(df)
        # NOTE - can pass expected schema and columns for output df here
        table: pa.Table = pa.Table.from_pandas(df, preserve_index=False)
        write_table(table, fn)

    if opt.HAVE_POLARS:

        @save_file.register  # type: ignore
        def _(fn: PathOrFile, df: opt.PlDataFrame):  # type: ignore # noqa: N805
            table: pa.Table = df.to_arrow()
            write_table(table, fn)

    save_file = staticmethod(save_file)  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

from pydantic import BaseModel, ConfigDict, computed_field

if TYPE_CHECKING:
    from arakawa.view import ViewVisitor


VV = TypeVar("VV", bound="ViewVisitor")


class BaseBlock(BaseModel):
    """Base Block class - subclassed by all Block types

    ..note:: The class is not used directly.
    """

    _tag: ClassVar[str]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field(alias="_tag")
    @property
    def computed_tag(self) -> str:
        return self._tag

    @computed_field
    @property
    def id(self) -> str | None:
        raise NotImplementedError("id property must be implemented in subclass")

    def _ipython_display_(self):
        """Display the block as a side effect within a Jupyter notebook"""
        from IPython.display import HTML, display

        from arakawa.ipython.environment import get_environment
        from arakawa.processors.api import stringify_report
        from arakawa.view import Blocks

        if get_environment().support_rich_display:
            html_str = stringify_report(Blocks(self))
            display(HTML(html_str))
        else:
            display(self.__str__())

    def _accept(self, visitor: VV) -> VV:
        # NOTE: use underscore prefix to avoid conflict
        visitor.visit(self)
        return visitor


class DataBlock(BaseBlock):
    """Abstract block that represents a leaf-node in the tree, e.g. a Plot or Table

    ..note:: This class is not used directly.
    """


BlockOrPrimitive = BaseBlock | Any  # TODO - expand
BlockList = list[BaseBlock]


def wrap_block(b: BlockOrPrimitive) -> BaseBlock:
    from .wrappers import convert_to_block

    # if isinstance(b, Page):
    #     raise DPClientError("Page objects can only be at the top-level")
    if not isinstance(b, BaseBlock):
        # import here as a very slow module due to nested imports
        # from ..files import convert
        return convert_to_block(b)

    return b



from __future__ import annotations

import json
import math
import secrets
from collections.abc import Sized
from numbers import Number
from typing import Annotated, Any

from pydantic import BeforeValidator


def gen_name() -> str:
    """Return a (safe) name for use in a Block"""
    return f"id-{secrets.token_urlsafe(8)}"


def convert_optional_str_or_number(v: Any) -> str | None:
    """
    Convert a value to a str for use as an ElementBuilder attribute
    - also handles None to a string for optional field values
    """
    if v is None:
        return v

    if isinstance(v, Sized) and len(v) == 0:
        return None

    if isinstance(v, str):
        return v

    if isinstance(v, Number) and not isinstance(v, bool):
        if math.isinf(v) and v > 0:  # type: ignore
            return "INF"

        if math.isinf(v) and v < 0:  # type: ignore
            return "-INF"

        if math.isnan(v):  # type: ignore
            return "NaN"

        return str(v)

    return json.dumps(v)


def convert_str_or_number(v: Any) -> str:
    """
    Convert a value to a str for use as an ElementBuilder attribute
    - also handles None to a string for optional field values
    """
    converted = convert_optional_str_or_number(v)
    if converted is None:
        raise ValueError("Value cannot be None or nullable")

    return converted


OptionalNumberStr = Annotated[
    str | None, BeforeValidator(convert_optional_str_or_number)
]
NumberStr = Annotated[str, BeforeValidator(convert_str_or_number)]



# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0

try:
    from ._version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"

# Public API re-exports
from .blocks import (  # noqa: F401
    HTML,
    Attachment,
    BigNumber,
    Block,
    ChoiceField,
    Code,
    ColorField,
    Compute,
    DataTable,
    DateField,
    DateTimeField,
    EmailField,
    Embed,
    Empty,
    FileField,
    Formula,
    GreatTables,
    Group,
    HiddenField,
    Media,
    MultiChoiceField,
    NetworkX,
    NumberBox,
    Page,
    PasswordField,
    Plot,
    RangeField,
    SearchField,
    Select,
    SelectType,
    Sigma,
    SwitchField,
    Table,
    TagsField,
    TelephoneField,
    Text,
    TextareaField,
    TextBox,
    TimeField,
    Toggle,
    URLField,
    wrap_block,
)
from .exceptions import ARError  # noqa: F401
from .processors import (  # noqa: F401
    FontChoice,
    Formatting,
    TextAlignment,
    Width,
    save_report,
    stringify_report,
)
from .view import Blocks, Report, View  # noqa: F401

from . import builtins  # noqa: F401 # isort: skip

X = wrap_block



from __future__ import annotations

import sys
from collections.abc import Callable
from typing import Any

import humps
from pydantic import Field

from arakawa import optional_libs as opt
from arakawa.exceptions import ARError

from .base import DataBlock
from .mixins import OptionalLabelMixin, OptionalNameMinx

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class LayoutSettings(TypedDict, total=False):
    adjust_sizes: bool | None
    barnes_hut_optimize: bool | None
    barnes_hut_theta: int | float | None
    edge_weight_influence: int | float | None
    gravity: int | float | None
    lin_log_mode: bool | None
    outbound_attraction_distribution: bool | None
    scaling_ratio: int | float | None
    slow_down: int | float | None
    strong_gravity_mode: bool | None


class Sigma(OptionalNameMinx, OptionalLabelMixin, DataBlock):
    _tag = "Sigma"

    data: dict[str, Any] = Field(...)
    width: int = Field(...)
    height: int = Field(...)
    layout_settings: LayoutSettings = Field(...)

    def __init__(
        self,
        data: dict[str, Any],
        width: int = 960,
        height: int = 540,
        layout_settings: LayoutSettings | None = None,
        name: str | None = None,
        label: str | None = None,
    ):
        """
        Args:
            data: A Sigma/Graphology graph data to attach
            width: A width of the graph
            height: A height of the graph
            layout_settings: see https://www.npmjs.com/package/graphology-layout-forceatlas2#settings (optional, each key should be snake_cased)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(
            data=data,
            layout_settings=humps.camelize(layout_settings or {}),
            name=name,
            label=label,
            width=width,
            height=height,
        )


def network_graph_to_sigma(G: opt.NXGraph, layout_function: Callable[[Any], Any]):  # noqa: N803
    data = {
        "options": {
            "type": "directed" if G.is_directed() else "undirected",
            "multi": bool(G.is_multigraph()),
            "allowSelfLoops": True,
        },
        "attributes": G.graph,
        "nodes": [
            {"key": node_id, "attributes": attributes}
            for node_id, attributes in dict(G.nodes(data=True)).items()
        ],
        "edges": [
            {"key": idx, "source": src, "target": tar, "attributes": attr}
            for idx, (src, tar, attr) in enumerate(list(G.edges(data=True)))
        ],
    }
    absence_of_layout_info = False
    for node_data in data["nodes"]:
        if "x" not in node_data["attributes"] or "y" not in node_data["attributes"]:
            absence_of_layout_info = True
            break

    if absence_of_layout_info:
        layout = layout_function(G)
        for node_data in data["nodes"]:
            node_data["attributes"]["x"] = layout[node_data["key"]][0]
            node_data["attributes"]["y"] = layout[node_data["key"]][1]

    return data


class NetworkX(Sigma):
    """
    The NetworkX block allows you to embed an NetworkX graph (rendered by Sigma.js) into your app.
    """

    def __init__(
        self,
        graph: opt.NXGraph,
        width: int = 960,
        height: int = 540,
        layout_settings: LayoutSettings | None = None,
        layout_function: Callable[[opt.NXGraph], Any] | None = None,
        name: str | None = None,
        label: str | None = None,
    ):
        """
        Args:
            graph: An NetworkX graph to attach
            width: A width of the graph
            height: A height of the graph
            layout_settings: see https://www.npmjs.com/package/graphology-layout-forceatlas2#settings (optional, each key should be snake_cased)
            layout_function: A layout function. Defaults to spring_layout.
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        if not opt.HAVE_NETWORKX:
            raise ARError(
                "NetworkX is not installed. Please install networkx to use this block."
            )

        super().__init__(
            data=network_graph_to_sigma(
                graph, layout_function=layout_function or opt.networkx.spring_layout
            ),
            layout_settings=layout_settings,
            name=name,
            label=label,
            height=height,
            width=width,
        )



from __future__ import annotations

from pydantic import Field, computed_field

from .base import DataBlock
from .mixins import NameMixin, OptionalLabelMixin


class ControlBlock(NameMixin, DataBlock):
    """
    Abstract block for all the control blocks.
    """

    @computed_field(alias="name")
    @property
    def name_(self) -> str:
        return self.name


class OptionalHelpMixin(DataBlock):
    help: str | None = Field(default=None)


class OptionalRequiredMixin(DataBlock):
    required: bool | None = Field(default=None)


class OptionalValidationMixin(DataBlock):
    validation: str | None = Field(default=None)


class OptionalStringInitialMixin(DataBlock):
    initial: str | None = Field(default=None)


class BaseDateTime(
    OptionalStringInitialMixin,
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    Abstract block for DateTime, Date and Time blocks.
    """

    def __init__(
        self,
        name: str,
        help: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            help (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            help=help,
            validation=validation,
        )


class DateTimeField(BaseDateTime):
    """
    DateTimeField allows you to add a datetime type input.
    """

    _tag = "DateTimeField"

    __init__ = BaseDateTime.__init__


class DateField(BaseDateTime):
    """
    DateField allows you to add a date type input.
    """

    _tag = "DateField"

    __init__ = BaseDateTime.__init__


class TimeField(BaseDateTime):
    """
    TimeField allows you to add a time type input.
    """

    _tag = "TimeField"

    __init__ = BaseDateTime.__init__


class FileField(
    OptionalStringInitialMixin,
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    FileField allows you to add a file type input.
    """

    _tag = "FileField"

    accept: str | None = Field(default=None)

    def __init__(
        self,
        name: str,
        accept: str | None = None,
        help: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            help (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
            accept (str | None, optional): An accept value. Defaults to None.
        """
        super().__init__(
            name=name,
            accept=accept,
            help=help,
            label=label,
            required=required,
            validation=validation,
        )


class MultiChoiceField(
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    MultiChoiceField allows you to have a multiple select type input.
    """

    _tag = "MultiChoiceField"

    initial: list[str] = Field(..., min_length=1)
    options: list[str] = Field(..., min_length=1)

    def __init__(
        self,
        name: str,
        initial: list[str],
        options: list[str],
        help: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            initial (list[str]): An initial value.
            options (list[str]): Options.
            help (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            initial=initial,
            options=options,
            help=help,
            label=label,
            required=required,
            validation=validation,
        )


class NumberBox(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    NumberBox allows you to add a number type input.
    """

    _tag = "NumberBox"

    initial: int | float | None = Field(default=None)

    def __init__(
        self,
        name: str,
        help: str | None = None,
        initial: int | float | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            help (str | None, optional): A help text. Defaults to None.
            initial (int | float | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            help=help,
            initial=initial,
            label=label,
            required=required,
            validation=validation,
        )


class RangeField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalLabelMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    """
    RangeField allows you to add a range type input.
    """

    _tag = "RangeField"

    initial: int | float = Field(...)
    min: int | float = Field(...)
    max: int | float = Field(...)
    step: int | float = Field(...)

    def __init__(
        self,
        name: str,
        min: int | float,
        max: int | float,
        step: int | float,
        help: str | None = None,
        initial: int | float | None = None,
        label: str | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): Name.
            min (int | float): A min value.
            max (int | float): A max value.
            step (int | float): A step value.
            help (str | None, optional): A help text. Defaults to None.
            initial (int | float | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        initial = initial or min
        super().__init__(
            name=name,
            help=help,
            initial=initial,
            label=label,
            max=max,
            min=min,
            step=step,
            validation=validation,
        )


class ChoiceField(
    OptionalValidationMixin,
    OptionalStringInitialMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    ChoiceField allows you to add a select type input.
    """

    _tag = "ChoiceField"

    options: list[str] = Field(..., min_length=1)  # type: ignore

    def __init__(
        self,
        name: str,
        options: list[str],
        help: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            options (list[str]): Options.
            help (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            options=options,
            help=help,
            initial=initial,
            label=label,
            required=required,
            validation=validation,
        )


class SwitchField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    SwitchField allows you to add a checkbox type input.
    """

    _tag = "SwitchField"

    initial: bool | None = Field(default=None)

    def __init__(
        self,
        name: str,
        help: str | None = None,
        initial: bool | None = None,
        label: str | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            help (str | None, optional): A help text. Defaults to None.
            initial (bool | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name, help=help, initial=initial, label=label, validation=validation
        )


class TagsField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    """
    TagsField allows you to add a multi select type input along with the free form input.
    """

    _tag = "TagsField"

    initial: list[str] = Field(..., min_length=1)  # type: ignore

    def __init__(
        self,
        name: str,
        initial: list[str],
        help: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            initial (list[str]): An initial value.
            help (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            initial=initial,
            help=help,
            label=label,
            required=required,
            validation=validation,
        )


class BaseTextField(
    OptionalStringInitialMixin,
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalLabelMixin,
    ControlBlock,
):
    caption: str | None = Field(default=None)
    help: str | None = Field(default=None)
    initial: str | None = Field(default=None)
    label: str | None = Field(default=None)
    required: bool | None = Field(default=None)
    validation: str | None = Field(default=None)

    def __init__(
        self,
        name: str,
        help: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (str): A name.
            help (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            validation=validation,
            help=help,
        )


class TextBox(BaseTextField):
    """
    TextBox allows you to add a text type input.
    """

    _tag = "TextBox"

    __init__ = BaseTextField.__init__


class URLField(BaseTextField):
    """
    URLField allows you to add a url type input.
    """

    _tag = "URLField"

    __init__ = BaseTextField.__init__


class EmailField(BaseTextField):
    """
    EmailField allows you to add an email type input.
    """

    _tag = "EmailField"

    __init__ = BaseTextField.__init__


class SearchField(BaseTextField):
    """
    SearchField allows you to add a search type input.
    """

    _tag = "SearchField"

    __init__ = BaseTextField.__init__


class TelephoneField(BaseTextField):
    """
    TelephoneField allows you to add a tel type input.
    """

    _tag = "TelephoneField"

    __init__ = BaseTextField.__init__


class PasswordField(BaseTextField):
    """
    PasswordField allows you to add a password type input.
    """

    _tag = "PasswordField"

    __init__ = BaseTextField.__init__


class TextareaField(BaseTextField):
    """
    TextareaField allows you to add a search type input.
    """

    _tag = "TextareaField"

    __init__ = BaseTextField.__init__


class HiddenField(ControlBlock):
    """
    HiddenField allows you to add a hidden type input.
    """

    _tag = "HiddenField"

    initial: str = Field(...)

    def __init__(
        self,
        name: str,
        initial: str,
    ):
        """
        Args:
            name (str): A name.
            initial (str): An initial value.
        """
        super().__init__(
            name=name,
            initial=initial,
        )


class ColorField(BaseTextField):
    """
    ColorField allows you to add a color picker.
    """

    _tag = "ColorField"

    __init__ = BaseTextField.__init__



from __future__ import annotations

from functools import partial
from typing import Annotated

from pydantic import (
    BeforeValidator,
    Field,
    computed_field,
    field_validator,
)

from arakawa.common.utils import is_valid_id

from .base import DataBlock


def truncate_optional_string(x: str | None, *, max_length: int) -> str | None:
    if x is None or len(x) <= max_length:
        return x

    return f"{x[: max_length - 3]}..."


class OptionalLabelMixin(DataBlock):
    label: Annotated[
        str | None, BeforeValidator(partial(truncate_optional_string, max_length=256))
    ] = Field(
        default=None,
        description="A label to display above the block",
        min_length=1,
        max_length=256,
    )


class OptionalCaptionMixin(DataBlock):
    caption: Annotated[
        str | None, BeforeValidator(partial(truncate_optional_string, max_length=512))
    ] = Field(
        default=None,
        description="A caption to display below the block",
        min_length=1,
        max_length=512,
    )


class OptionalNameMinx(DataBlock):
    # TODO: deprecate name and use id only
    name: str | None = Field(default=None, deprecated=True, exclude=True)

    @computed_field
    @property
    def id(self) -> str | None:
        return self.name

    @field_validator("name", mode="after")
    @classmethod
    def _validate_name(cls, v: str | None):
        if v and not is_valid_id(v):
            raise ValueError(f"Invalid name '{v}' for block")

        return v


class NameMixin(DataBlock):
    # TODO: deprecate name and use id only
    name: str = Field(..., deprecated=True, exclude=True)

    @computed_field
    @property
    def id(self) -> str | None:
        return self.name

    @field_validator("name", mode="after")
    @classmethod
    def _validate_name(cls, v: str):
        if not is_valid_id(v):
            raise ValueError(f"Invalid name '{v}' for block")

        return v
